#include <stdio.h>         /* Teletextplatine mit 80535 */
#include <stdlib.h>
#include <ioumfunc.h>
#include <string.h>
#include <ctype.h>
#include <io51.h>          /* Header fr 8051 */
#include <delay.h>

#define byte unsigned char

#define CLOCK P1.1  /* Pin an dem der Clock anliegen sollen */
#define DATEN P1.0  /* Pin an den die Daten anliegen sollen */
#define AVDAT P1.0  /* Pin fr AVDATEN */
#define DSTRB P1.2  /* Strobe fr AVDATEN */
#define IICBD 0xBC  /* Geschwindigkeit des Clocks war BC */
#define REL_RX 0xFFF /* Blinkgeschwindigkeit LED bei RX */
#define REL_TX 0x3FF /* Blinkgeschwindigkeit LED bei TX */
#define RXUG 38001   /* Untergrenze 70cm Band RX */
#define RXOG 38799   /* Obergrenze  70cm Band RX */
#define TXUG 430000000   /* Untergrenze 70cm Band TX */
#define TXOG 440000000   /* Obergrenze  70cm Band TX */

#define XRAM ((char *) 0x01C000)
#define EEPR ((char *) 0x014000)
#define EEDEL 1000
#define INPT ((char *) 0x018000)  /* Adresse Eingangslatch */
#define OUT0 ((char *) 0x010000)  /* Adresse Ausgangslatch 0 */
#define OUT1 ((char *) 0x010001)  /* Adresse Ausgangslatch 1 */
#define RX   ((char *) 0x014020)  /* EE-Adresse RX-QRG Einschaltkanal*/
#define TX   ((char *) 0x014028)  /* EE-Adresse TX-QRG Einschaltkanal*/
#define RX1  ((char *) 0x014030)  /* EE-Adresse RX-QRG Kanal 1*/
#define TX1  ((char *) 0x014038)  /* EE-Adresse TX-QRG Kanal 1*/
#define RX2  ((char *) 0x014040)  /* EE-Adresse RX-QRG Kanal 2*/
#define TX2  ((char *) 0x014048)  /* EE-Adresse TX-QRG Kanal 2*/
#define RX3  ((char *) 0x014050)  /* EE-Adresse RX-QRG Kanal 3*/
#define TX3  ((char *) 0x014058)  /* EE-Adresse TX-QRG Kanal 3*/
#define RX4  ((char *) 0x014060)  /* EE-Adresse RX-QRG Kanal 4*/
#define TX4  ((char *) 0x014068)  /* EE-Adresse TX-QRG Kanal 4*/
#define RX5  ((char *) 0x014070)  /* EE-Adresse RX-QRG Kanal 5*/
#define TX5  ((char *) 0x014078)  /* EE-Adresse TX-QRG Kanal 5*/
#define RX6  ((char *) 0x014080)  /* EE-Adresse RX-QRG Kanal 6*/
#define TX6  ((char *) 0x014088)  /* EE-Adresse TX-QRG Kanal 6*/
#define RX7  ((char *) 0x014090)  /* EE-Adresse RX-QRG Kanal 7*/
#define TX7  ((char *) 0x014098)  /* EE-Adresse TX-QRG Kanal 7*/
#define DX   ((char *) 0x014100)  /* EE-Adresse DX-Abstand */
#define CHAN ((char *) 0x014108)  /* Kanalspeicher */


/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   RAM Adresse 0xC000 - 0xFFFF
   EEPROM-Adresse 0x4000
   IN-LATCH-Adresse 0x8000
   OUT-LATCH-Adresse 0x0000 und 0x0001
   ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

char *power = "\x00\x00\x80\xC0";                        /* Sendeleistungen */
xdata char avstat[6],serein,einsp[255],mem,tast_prell,anzfreqstr[80];
xdata char rx[8],tx[8];
xdata unsigned long ledtimer,status;
xdata unsigned long tx_vco_mhz;
xdata long rpt,shift;
xdata unsigned int pulstimer;
bit ausles,trbit,bitptt,bitlptt,bitled,amateur,mem_up,duplex,lowpara;

void inadresse(char *adrstr);
void outadresse(char *adrstr);
void pllset(char l);
char freqeingabe(char *freq);
unsigned long freqbin(char *str);
void ladetxqrg();
void laderxqrg();
void system(char *str);
void memory(char memo);
void txvco(char *str);
void freqanzeige(char kanal);
void freqasc(unsigned long wert);
void kanalanz(char i);

void main(void)
{
 OUT1[0] = 0x50;                                          /* Koaxrelais aus */
 P1 = 0xFB;P3 = 0xFF;
 TMOD = 0x22; TCON = 0x40; TH1 = 0xFD;      	       /* Timer Einstellung */
 ET0 = 0; ET1 = 0; ES = 1; EAL = 1;    		 /* Interrupt Einstellungen */

 ausles = 0;ledtimer = 0;status = REL_RX;pulstimer = 0;amateur = 0;mem = 0;
 tast_prell = 0;lowpara = 0;

 SCON = 0x74; PCON = 0x00;                /* Serielle Einstellung 9600 Baud */

 /* Einschalttext */
 putstring("Storno Motronic 6000 C oder PRM6663D15A\rAFU-Firmware by OE5PON V7.3 (C)\r8 Speicherpltze, Vollduplex\r\r");

 /* Duplexabstand berechnen */
 shift = (long)(ascbin(DX,5));
 rpt = shift * (-1000);
 tx_vco_mhz = (unsigned long)(rpt + 45000000);

 DX[6] = 0;delayi(EEDEL);
 if(CHAN[0] > 7)
  {
   CHAN[0] = CHAN[0]&7;
   delayi(EEDEL);  /* Einschaltkanal */
  }

 /* Wenn Pin 21 low dann lade Testfrequenz beim Einschalten */
 if((INPT[0]&0x20) == 0)
  {
   txvco("-10000");
   putstring("Testfrequenz wird geladen\r");
   freqeingabe("R0 4340000");
   freqeingabe("T0 4340000");
  }
 else
  {
   putstring("Vollduplexabstand = ");
   putstring(DX);
   putstring(" kHz");
   putc('\r');
   freqanzeige(CHAN[0]);
   memory(CHAN[0]);
   laderxqrg();
  }


 bitlptt = 1;

 for(;;)
  {
   if((++ledtimer) == status)                                /* LED blinken */
    {
     P1.6 = bitled = ~bitled;
     ledtimer = 0;
    }

   bitptt = ~(INPT[0]&1);                                   /* PTT einlesen */
   if(bitlptt != bitptt)                           /* hat sich PTT gendert */
    {
     if(bitptt)                                    /* TX wenn PTT vorhanden */
      {
       OUT1[0] = 0x00;                            /* Koaxrelais einschalten */
       if(duplex)
	{
	 avstat[0] = tx[4]|0x10;pllset(1);        /* nur Sender einschalten */
	}
       else                                          /* auch Frequenz laden */
	ladetxqrg();
       status = REL_TX;
       ledtimer = 0;
      }
     else                                              /* RX wenn keine PTT */
      {
       if(duplex)
	{
	 avstat[0] = 0x00;pllset(1);
	}
       else
	laderxqrg();
       OUT1[0] = 0x50;
       status = REL_RX;
       ledtimer = 0;
      }
    }
   bitlptt = bitptt;

  /* Sperre bei Erstprogrammierung keine Kanalumschaltung! */
  if((INPT[0]&0x20) == 0x20)
   lowpara = 1;

  /* Kanalweiterschaltung */
  if((INPT[0]&0x20) == 0  && lowpara == 1)
   {
    if(tast_prell>=250)
     mem_up = 1;
    else
     ++tast_prell;
   }
  else
   {
    tast_prell = 0;
    if(mem_up)
     {
      mem = (++CHAN[0])&0x07;
/*    if (mem == 0)
       mem++; */
      memory(mem);
      laderxqrg();
      mem_up = 0;
     }
   }

   if (ausles)           /* wenn Serielle Eingabe erfolgte, erst nach Enter */
    {
     ausles = 0;
     switch(toupper(einsp[0]))      /* erstes Zeichen nach Eingabe */
      {
       case '?':putstring("Fk   Frequenzanzeige(k=Kanal 0-7)\rT    Senden\rR    Empfangen\rFTk  <Sendefrequenz(MHz)>(k=Kanal 1-7)\rFRk  <Empfangsfrequenz(MHz)>(k=Kanal 1-7)\rMk   Speicherkanal(k = Kanal 0-7)\rPk x Sendeleistung(k = 0-7;x = 1-3)\r\r");break;
       case 'D':if(einsp[1] != ' ') einsp[2] = '?';txvco(&einsp[2]);break;
       case 'I':inadresse(&einsp[2]);break;
       case 'O':outadresse(&einsp[2]);break;
       case 'X':system(&einsp[2]);break;
       case 'F':freqeingabe(&einsp[1]);break;
       case 'T':OUT1[0] = 0x00;ladetxqrg();status = REL_TX;ledtimer = 0;putstring("TX OK\r");break;
       case 'R':laderxqrg();OUT1[0] = 0x50;status = REL_RX;ledtimer = 0;putstring("RX OK\r");break;
       case 'P':TX[4+((einsp[1]&7)*16)] = power[(einsp[3]&3)];delayi(EEDEL);
/*		if(mem == (einsp[1]&7)) {TX[4] = power[(einsp[3]&3)];delayi(EEDEL);}*/
		freqanzeige(einsp[1]&7);
/*		if(bitptt || duplex) ladetxqrg();putstring("OK\r");*/
       case 'M':memory(einsp[1]);laderxqrg();break;
       case 'S':avstat[0] = tx[4]|0x10;pllset(1);OUT1[0] = 0x00;status = REL_TX;ledtimer = 0;putstring("TX OK\r");break;
       case 'E':avstat[0] = 0x00;pllset(1);OUT1[0] = 0x50;status = REL_RX;ledtimer = 0;putstring("RX OK\r");break;
      }
    }
  }
  if(0)
   {
    puthex(0);avset();delayl(0);
   }
}

void txvco(char *str)
{
 if(str[0] != '+' && str[0] != '-' && str[0] != '0' && str[0] != '1')
  {
   putstring("Eingabe des Vollduplexabstand: (Von Hardware abhngig!)\r");
   putstring("Beispiel:  D +07600 oder D -09600\r\r");
  }
 else
  {
   shift = (long)(ascbin(str,5));
   if(shift >= -10000 && shift <= 10000)
    {
     DX[0] = str[0];delayi(EEDEL);  /* optionales Vorzeichen */
     DX[1] = str[1];delayi(EEDEL);
     DX[2] = str[2];delayi(EEDEL);
     DX[3] = str[3];delayi(EEDEL);
     DX[4] = str[4];delayi(EEDEL);
     DX[5] = str[5];delayi(EEDEL);
     DX[6] = str[6];delayi(EEDEL);
    }
   else
    {
     shift = (long)(ascbin(DX,5));
     putstring("Neu eingegebener Duplexabstand nicht im mglichen Bereich!\r\r");
    }
   rpt = shift * (-1000);
   tx_vco_mhz = (unsigned long)(rpt + 45000000);
  }
 putstring("Programmierter Vollduplexabstand = ");putstring(DX);putstring(" kHz");putc('\r');
}

void memory(char memo)
{
 char i;
 if(memo == '?')
  {
   putstring("Kanalumschaltung: M<Kanal><Enter>\r");
   putstring("Aktiver Speicherkanal = ");putc(0x30|CHAN[0]);putc('\r');
  }
 else
  {
   memo &= 0x07;
      rx[0] = RX[0+memo*16];delayi(EEDEL);
      rx[1] = RX[1+memo*16];delayi(EEDEL);
      rx[2] = RX[2+memo*16];delayi(EEDEL);
      rx[3] = RX[3+memo*16];delayi(EEDEL);
      rx[4] = RX[4+memo*16];delayi(EEDEL);
      rx[0] = TX[0+memo*16];delayi(EEDEL);
      tx[1] = TX[1+memo*16];delayi(EEDEL);
      tx[2] = TX[2+memo*16];delayi(EEDEL);
      tx[3] = TX[3+memo*16];delayi(EEDEL);
      tx[4] = TX[4+memo*16];delayi(EEDEL);
      if(memo != CHAN[0])
       {
	CHAN[0] = memo;delayi(EEDEL);
       }
      duplex = 1;
      for(i=0;i<=3;++i)
       {
	if(rx[i] != tx[i])
	 duplex = 0;
       }
      putstring("Aktiver Speicherkanal = ");putc(0x30|CHAN[0]);
      if(duplex)
       putstring(" (Vollduplex Kanal)\r");
      else
       putc('\r');
  }
}

void laderxqrg()   /* ldt die PLL mit der RX QRG und schaltet auf RX */
{
 avstat[0] = rx[0];
 avstat[1] = rx[1];
 avstat[2] = rx[2];
 avstat[3] = rx[3];
 avstat[4] = 0x01;                                /* Sender immer aus */
 pllset(5);
}

void ladetxqrg()   /* ldt die PLL mit der TX QRG und schaltet auf TX */
{                                 /* Sendeleistung bleibt unverndert */
 avstat[0] = tx[0];
 avstat[1] = tx[1];
 avstat[2] = tx[2];
 avstat[3] = tx[3];
 avstat[4] = tx[4]|0x11;                    /* TX[4]=Power, 0x10 Sender ein */
 pllset(5);
}

void inadresse(char *adrstr)
{
 unsigned char hadr,ladr;
 unsigned int adr;
 hadr = hexbin(&adrstr[0]);
 ladr = hexbin(&adrstr[2]);
 adr = (hadr << 8) | ladr;
 puthex(OUT0[adr]);
}

void outadresse(char *adrstr)
{
 unsigned char hadr,ladr,wert;
 unsigned int adr;
 hadr = hexbin(&adrstr[0]);
 ladr = hexbin(&adrstr[2]);
 wert = hexbin(&adrstr[4]);
 adr = (hadr << 8) | ladr;
 OUT0[adr] = wert;
 delayi(EEDEL);
}

/* Telegramm fr Funkgertesteuerung
   PP0T 0001  RRRR RRRR  RRRR RRNN  NNNN NNNN  AAAA AAA1
   P = Leistungsstufe
   T = Sender Ein/Aus
   R = Rasterteiler
   N = NTeiler
   A = ATeiler
*/

char freqeingabe(char *freq)   /* fixer Raster !!!! */
{
 unsigned long rxfreq,nteiler,ateiler,frequenz,teiler;
 char n,err;
 n = freq[1]&0x07;
 err = 1;
 if(freq[2] == '4')
  frequenz = freqbin(&freq[2]);
 else
  frequenz = freqbin(&freq[3]);

 if(freq[0] == 0)
  {
   freqanzeige(8);
   err = 0;
  }

 if((freq[0]&0xF8) == 0x30)
  {
   freqanzeige(freq[0]&0x7);
   err = 0;
  }

 if((toupper(freq[0])) == 'R')
  {
   teiler = (frequenz + 45000000) / 12500;
   if((teiler >= RXUG && teiler<= RXOG) || amateur)
     {
       nteiler = teiler / 64;
       ateiler = teiler - (nteiler * 64); /* Der Rest macht den Unterschied */
       avstat[0] = 0x14;
       avstat[1] = 0x20; avstat[1]|= (nteiler >> 8)&0x03;
       avstat[2] = nteiler&0xFF;
       avstat[3] = (ateiler << 1)|0x01;     /* Raster wird immer mitgeladen */
       avstat[4] = 0x00;
       RX[0+n*16]=avstat[0];delayi(EEDEL);
       RX[1+n*16]=avstat[1];delayi(EEDEL);
       RX[2+n*16]=avstat[2];delayi(EEDEL);
       RX[3+n*16]=avstat[3];delayi(EEDEL);
       RX[4+n*16]=avstat[4];delayi(EEDEL);
       freqanzeige(n);
       memory(n);
       if(!bitptt || duplex)         /* bei Empfang und Duplex sofort laden */
	laderxqrg();
       err = 0;
     }
   else
     {
       putstring("Fehler! Frequenzeingabe auerhalb vom Amateurband\r");return(1);
       err = 2;
     }
  }

 if((toupper(freq[0])) == 'T')
  {
   teiler = (frequenz + tx_vco_mhz) / 12500;
    if((frequenz > TXUG && frequenz < TXOG) || amateur)
     {
       nteiler = teiler / 64;
       ateiler = teiler - (nteiler * 64);
       avstat[0] = 0x14;
       avstat[1] = 0x20; avstat[1]|= (nteiler >> 8)&0x03;
       avstat[2] = nteiler&0xFF;
       avstat[3] = (ateiler << 1)|0x01;     /* Raster wird immer mitgeladen */
       avstat[4] = tx[4]&0xC0;               /* nur Sendeleistung speichern */
       TX[0+n*16]=avstat[0];delayi(EEDEL);
       TX[1+n*16]=avstat[1];delayi(EEDEL);
       TX[2+n*16]=avstat[2];delayi(EEDEL);
       TX[3+n*16]=avstat[3];delayi(EEDEL);
       TX[4+n*16]=avstat[4];delayi(EEDEL);
       freqanzeige(n);
       memory(n);
       if(bitptt || duplex)           /* bei Senden und Duplex sofort laden */
	ladetxqrg();
       err = 0;
     }
   else
     {
       putstring("Fehler! Frequenzeingabe auerhalb vom Amateurband\r");return(1);
       err = 2;
     }
  }

 if((toupper(freq[0])) == '?' || err == 1)
  {
   putstring("Hilfe: Frequenzeingabe\r");
   putstring("F<Kanal><Enter> Frequenzanzeige\r");
   putstring("FT<Kanal> <QRG><Enter> Tx-Frequenzeingabe\r");
   putstring("FR<Kanal> <QRG><Enter> Rx-Frequenzeingabe\r");
   putstring("Frequenzeingabe im 12,5kHz Raster mglich!\r");
   putstring("Nullen am Ende der QRG brauchen nicht eingegeben werden!\r\r");
   err = 0;
  }

 return(0);
}

void freqanzeige(char kanal)
{
 char i;
 if(kanal == 8)
  {
   for(i=0;i<=7;i++)
    kanalanz(i);
  }
 else
  kanalanz(kanal);
}

void kanalanz(char i)
{
 unsigned long anzteiler,anzfreq;

 anzteiler = (unsigned long)(((RX[2+i*16] + ((RX[1+i*16]&3) << 8)) << 6) + (RX[3+i*16] >> 1));
 anzfreq = ((anzteiler&0xFFFF) * 12500) - 45000000;
 freqasc(anzfreq);
 anzfreqstr[0] = 'F';
 anzfreqstr[1] = 'R';
 anzfreqstr[2] = 0x30|i;
 anzfreqstr[3] = ' ';
 anzfreqstr[13] = '\r';
 anzfreqstr[14] = 0;
 putstring(anzfreqstr);

 anzteiler = (unsigned long)(((TX[2+i*16] + ((TX[1+i*16]&3) << 8)) << 6) + (TX[3+i*16] >> 1));
 anzfreq = ((anzteiler&0xFFFF) * 12500) - tx_vco_mhz;
 freqasc(anzfreq);
 anzfreqstr[0] = 'F';
 anzfreqstr[1] = 'T';
 anzfreqstr[2] = 0x30|i;
 anzfreqstr[3] = ' ';
 anzfreqstr[13] = '\r';
 anzfreqstr[14] = 0;
 putstring(anzfreqstr);

 anzfreqstr[0] = 'P';
 anzfreqstr[1] = 0x30|i;
 anzfreqstr[2] = ' ';
 switch(TX[4+i*16]&0xC0)
  {
   case 0x00:anzfreqstr[3] = '1';break;
   case 0x40:anzfreqstr[3] = '1';break;
   case 0x80:anzfreqstr[3] = '2';break;
   case 0xC0:anzfreqstr[3] = '3';break;
  }
 anzfreqstr[4] = '\r';
 anzfreqstr[5] = 0;
 putstring(anzfreqstr);
}


void freqasc(unsigned long freq)
{
  unsigned long wert;
  anzfreqstr[13] = 0;
  wert=freq/100;
  anzfreqstr[12] = (char)((wert%10)|0x30)&0x003F;
  wert=freq/1000;
  anzfreqstr[11] = ',';
  anzfreqstr[10] = (char)((wert%10)|0x30)&0x003F;
  wert=freq/10000;
  anzfreqstr[9] = (char)((wert%10)|0x30)&0x003F;
  wert=freq/100000;
  anzfreqstr[8] = (char)((wert%10)|0x30)&0x003F;
  wert=freq/1000000;
  anzfreqstr[7] = '.';
  anzfreqstr[6] = (char)((wert%10)|0x30)&0x003F;
  wert=freq/10000000;
  anzfreqstr[5] = (char)((wert%10)|0x30)&0x003F;
  wert=freq/100000000;
  anzfreqstr[4] = (char)(wert|0x30)&0x003F;
  anzfreqstr[3] = ' ';
  anzfreqstr[1] = 'T';
  anzfreqstr[0] = 'F';
}

void system(char *str)
{
 if(!strcmp(str,"Amateur"))
  {
   amateur = 0;
   putstring("Gert im Amateurmodus\r");
  }

 if(!strcmp(str,"Service"))
  {
   amateur = 1;
   putstring("Achtung! Gert im Servicemodus, nur zum Abgleich am Meplatz verwenden!\r");
  }
}

unsigned long freqbin(char *str)  /* ASCII-dezimal in Binr */
{
 char i;
 unsigned long bin=0,mult=100000000;
 strcat(str,"000000000");
   for (i=0;i<9;i++)
    {
     if(isdigit(str[i]))
      {
       bin+= (str[i]&0x0F)*mult;
       mult/=10;
      }
    }
 return bin;
}

void pllset(char l)
{
 data signed char i;
 DSTRB = 0;
 AVDAT = 0;
 CLOCK = 1;
 for(i=0;i<l;i++)
  {
     CLOCK = 1;
     _opc(0);_opc(0);_opc(0);_opc(0);_opc(0);
     CLOCK = 0;
     if(avstat[i]&0x80)
      AVDAT = 1;
     else
      AVDAT = 0;
     CLOCK = 1;
     _opc(0);_opc(0);_opc(0);_opc(0);_opc(0);
     CLOCK = 0;
     if(avstat[i]&0x40)
      AVDAT = 1;
     else
      AVDAT = 0;
     CLOCK = 1;
     _opc(0);_opc(0);_opc(0);_opc(0);_opc(0);
     CLOCK = 0;
     if(avstat[i]&0x20)
      AVDAT = 1;
     else
      AVDAT = 0;
     CLOCK = 1;
     _opc(0);_opc(0);_opc(0);_opc(0);_opc(0);
     CLOCK = 0;
     if(avstat[i]&0x10)
      AVDAT = 1;
     else
      AVDAT = 0;
     CLOCK = 1;
     _opc(0);_opc(0);_opc(0);_opc(0);_opc(0);
     CLOCK = 0;
     if(avstat[i]&0x08)
      AVDAT = 1;
     else
      AVDAT = 0;
     CLOCK = 1;
     _opc(0);_opc(0);_opc(0);_opc(0);_opc(0);
     CLOCK = 0;
     if(avstat[i]&0x04)
      AVDAT = 1;
     else
      AVDAT = 0;
     CLOCK = 1;
     _opc(0);_opc(0);_opc(0);_opc(0);_opc(0);
     CLOCK = 0;
     if(avstat[i]&0x02)
      AVDAT = 1;
     else
      AVDAT = 0;
     CLOCK = 1;
     _opc(0);_opc(0);_opc(0);_opc(0);_opc(0);
     CLOCK = 0;
     if(avstat[i]&0x01)
      AVDAT = 1;
     else
      AVDAT = 0;
  }
  CLOCK = 1;
  CLOCK = 0;
  _opc(0);_opc(0);_opc(0);_opc(0);_opc(0);
  DSTRB = 1;
  _opc(0);_opc(0);_opc(0);_opc(0);_opc(0);
  DSTRB = 0;
  _opc(0);_opc(0);_opc(0);_opc(0);_opc(0);
  AVDAT = 1;
  CLOCK = 1;
}

interrupt [0x0B] void T0_int (void)      /* Timer 0 Overflow */
{
 _opc(0xB2);_opc(0x91); /*   P1.1  Clock invertieren */
}

interrupt [0x23] void SCON_int (void)  /* Ser. Eing. Interrupt 9600Bd. */
{
 trbit = TR0;TR0 = 0;RI = 0;
 if ((SBUF != 0x0D)&(serein < 254))
  {
   if (SBUF == 0x0E)
    einsp[serein++] = 0x0D;
   else
    einsp[serein++] = SBUF;
  }
 else
  {
   einsp[serein] = 0;                   /* Stringende */
   ausles = 1;                          /* Stringende */
   serein = 0;
  }
  TR0 = trbit;
}
